@@ -9,7 +9,10 @@
 
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE MagicHash #-}
+{-# LANGUAGE MultiWayIf #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE QuasiQuotes #-}
 {-# LANGUAGE TemplateHaskell #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
@@ -33,6 +36,8 @@ where
 
 -- External Modules
 import           Control.Lens                ((^.), (%~), (&), (%=), (.~), view, makeLenses)
+import           Control.Applicative         ((<|>))
+import           Control.Monad.Extra         (ifM, andM)
 import           Control.Monad.RWS.Strict    (RWS)
 import qualified Control.Monad.RWS.Strict    as RWS
 import           Data.Bifunctor              (second)
@@ -47,41 +52,66 @@ import           Data.Text                   (Text, pack)
 import qualified Data.Text                   as Text
 import           Data.Text.Encoding          (decodeUtf8)
 import qualified Data.Traversable            as T
+import           Data.String.Interpolate     (__i)
 import qualified Text.Read                   as Text
+#if MIN_VERSION_ghc(9,4,0)
+import           Data.Primitive.ByteArray    (ByteArray(ByteArray))
+import qualified GHC.Data.Strict             as GHC
+import           GHC.Num.Integer             (integerToBigNatClamp#)
+#endif
+#if MIN_VERSION_ghc(9,6,0)
+import           Language.Haskell.Syntax.Basic (FieldLabelString (..))
+#endif
 
 -- GHC API
+#if MIN_VERSION_ghc(9,4,0)
+import GHC.Core.Reduction (Reduction(Reduction))
+#endif
 #if MIN_VERSION_ghc(9,0,0)
+import GHC.Builtin.Types (falseDataCon)
 import GHC.Core.Coercion.Axiom
   (CoAxiom (co_ax_branches), CoAxBranch (cab_lhs,cab_rhs), fromBranches)
 import GHC.Core.Coercion (Role (Nominal), coercionType, coercionKind)
 import GHC.Core.FVs  (exprSomeFreeVars)
 import GHC.Core
-  (AltCon (..), Bind (..), CoreExpr, Expr (..), Unfolding (..), Tickish (..),
+  (AltCon (..), Bind (..), CoreExpr, Expr (..), Unfolding (..),
+#if MIN_VERSION_ghc(9,2,0)
+   Alt(..),
+#else
+   Tickish (..),
+#endif
    collectArgs, rhssOfAlts, unfoldingTemplate)
+#if MIN_VERSION_ghc(9,2,0)
+import GHC.Types.Tickish (GenTickish (..))
+#endif
 import GHC.Core.DataCon
   (DataCon, dataConExTyCoVars, dataConName, dataConRepArgTys, dataConTag,
    dataConTyCon, dataConUnivTyVars, dataConWorkId, dataConFieldLabels, flLabel,
    HsImplBang(..), dataConImplBangs)
-import GHC.Driver.Session (unsafeGlobalDynFlags)
 import GHC.Core.FamInstEnv
   (FamInst (..), FamInstEnvs, familyInstances, normaliseType, emptyFamInstEnvs)
 import GHC.Data.FastString (unpackFS, bytesFS)
 import GHC.Types.Id (isDataConId_maybe)
 import GHC.Types.Id.Info (IdDetails (..), unfoldingInfo)
-import GHC.Types.Literal (Literal (..), LitNumType (..))
+import GHC.Types.Literal (Literal (..), LitNumType (..), literalType)
 import GHC.Unit.Module (moduleName, moduleNameString)
 import GHC.Types.Name
   (Name, nameModule_maybe, nameOccName, nameUnique, getSrcSpan)
 import GHC.Builtin.Names  (integerTyConKey, naturalTyConKey)
 import GHC.Types.Name.Occurrence (occNameString)
-import GHC.Utils.Outputable (showPpr)
 import GHC.Data.Pair (Pair (..))
 import GHC.Types.SrcLoc (SrcSpan (..), isGoodSrcSpan)
 import GHC.Core.TyCon
   (AlgTyConRhs (..), TyCon, tyConName, algTyConRhs, isAlgTyCon, isFamilyTyCon,
-   isFunTyCon, isNewTyCon, isPrimTyCon, isTupleTyCon,
+   isNewTyCon, isPrimTyCon, isTupleTyCon,
    isClosedSynFamilyTyConWithAxiom_maybe, expandSynTyCon_maybe, tyConArity,
    tyConDataCons, tyConKind, tyConName, tyConUnique, isClassTyCon, isPromotedDataCon_maybe)
+#if MIN_VERSION_ghc(9,6,0)
+import GHC.Core.TyCon (ExpandSynResult (..))
+import GHC.Core.Type (tyConAppFunTy_maybe)
+#else
+import GHC.Core.TyCon (isFunTyCon)
+#endif
 import GHC.Core.Type (mkTvSubstPrs, substTy, coreView)
 import GHC.Core.TyCo.Rep (Coercion (..), TyLit (..), Type (..), scaledThing)
 import GHC.Types.Unique (Uniquable (..), Unique, getKey, hasKey)
@@ -97,6 +127,7 @@ import CoreFVs    (exprSomeFreeVars)
 import CoreSyn
   (AltCon (..), Bind (..), CoreExpr, Expr (..), Unfolding (..), Tickish (..),
    collectArgs, rhssOfAlts, unfoldingTemplate)
+import TysWiredIn (falseDataCon)
 import DataCon    (DataCon, HsImplBang(..),
 #if MIN_VERSION_ghc(8,8,0)
                    dataConExTyCoVars,
@@ -107,7 +138,6 @@ import DataCon    (DataCon, HsImplBang(..),
                    dataConTag, dataConTyCon,
                    dataConUnivTyVars, dataConWorkId,
                    dataConFieldLabels, flLabel, dataConImplBangs)
-import DynFlags   (unsafeGlobalDynFlags)
 import FamInstEnv (FamInst (..), FamInstEnvs,
                    familyInstances, normaliseType, emptyFamInstEnvs)
 
@@ -120,12 +150,14 @@ import FastString (unpackFS, fastStringToByteString)
 import Id         (isDataConId_maybe)
 import IdInfo     (IdDetails (..), unfoldingInfo)
 import Literal    (Literal (..), LitNumType (..))
+#if MIN_VERSION_ghc(8,8,0)
+import Literal    (literalType)
+#endif
 import Module     (moduleName, moduleNameString)
 import Name       (Name, nameModule_maybe,
                    nameOccName, nameUnique, getSrcSpan)
 import PrelNames  (integerTyConKey, naturalTyConKey)
 import OccName    (occNameString)
-import Outputable (showPpr)
 import Pair       (Pair (..))
 import SrcLoc     (SrcSpan (..), isGoodSrcSpan)
 import TyCon      (AlgTyConRhs (..), TyCon, tyConName,
@@ -153,18 +185,21 @@ import VarSet     (isEmptyVarSet)
 
 -- Local imports
 import           Clash.Annotations.Primitive (extractPrim)
+import           Clash.Annotations.SynthesisAttributes (Annotate, Attr(..))
 import qualified Clash.Core.DataCon          as C
 import qualified Clash.Core.Literal          as C
 import qualified Clash.Core.Name             as C
+import qualified Clash.Core.Pretty           as C
 import qualified Clash.Core.Term             as C
 import qualified Clash.Core.TyCon            as C
 import qualified Clash.Core.Type             as C
 import qualified Clash.Core.Util             as C (undefinedTy, undefinedXPrims)
 import qualified Clash.Core.Var              as C
+import qualified Clash.Data.UniqMap          as C
 import           Clash.Normalize.Primitives  as C
-import           Clash.Primitives.Types
-import qualified Clash.Unique                as C
+import           Clash.Primitives.Types      hiding (name)
 import           Clash.Util
+import           Clash.GHC.Util
 
 instance Hashable Name where
   hashWithSalt s = hashWithSalt s . getKey . nameUnique
@@ -186,7 +221,7 @@ data GHC2CoreEnv
 makeLenses ''GHC2CoreEnv
 
 emptyGHC2CoreState :: GHC2CoreState
-emptyGHC2CoreState = GHC2CoreState C.emptyUniqMap HashMap.empty
+emptyGHC2CoreState = GHC2CoreState mempty HashMap.empty
 
 newtype SrcSpanRB = SrcSpanRB {unSrcSpanRB :: SrcSpan}
 
@@ -208,13 +243,13 @@ makeAllTyCons
 makeAllTyCons hm fiEnvs = go hm hm
   where
     go old new
-        | C.nullUniqMap (new ^. tyConMap) = C.emptyUniqMap
-        | otherwise                       = tcm `C.unionUniqMap` tcm'
+        | C.null (new ^. tyConMap) = mempty
+        | otherwise                = tcm <> tcm'
       where
         (tcm,old', _) = RWS.runRWS (T.mapM makeTyCon (new ^. tyConMap))
                                    (GHC2CoreEnv noSrcSpan fiEnvs)
                                    old
-        tcm'          = go old' (old' & tyConMap %~ (`C.differenceUniqMap` (old ^. tyConMap)))
+        tcm'          = go old' (old' & tyConMap %~ (`C.difference` (old ^. tyConMap)))
 
 makeTyCon :: TyCon
           -> C2C C.TyCon
@@ -270,16 +305,19 @@ makeTyCon tc = tycon
         mkTupleTyCon = do
           tcName <- coreToName tyConName tyConUnique qualifiedNameString tc
           tcKind <- coreToType (tyConKind tc)
-          tcDc   <- fmap (C.DataTyCon . (:[])) . coreToDataCon . head . tyConDataCons $ tc
-          return
-            C.AlgTyCon
-            { C.tyConUniq   = C.nameUniq tcName
-            , C.tyConName   = tcName
-            , C.tyConKind   = tcKind
-            , C.tyConArity  = tcArity
-            , C.algTcRhs    = tcDc
-            , C.isClassTc   = isClassTyCon tc
-            }
+          case tyConDataCons tc of
+            dc:_ -> do
+              tcDc   <- fmap (C.DataTyCon . (:[])) (coreToDataCon dc)
+              return
+                C.AlgTyCon
+                { C.tyConUniq   = C.nameUniq tcName
+                , C.tyConName   = tcName
+                , C.tyConKind   = tcKind
+                , C.tyConArity  = tcArity
+                , C.algTcRhs    = tcDc
+                , C.isClassTc   = isClassTyCon tc
+                }
+            _ -> error "impossible"
 
         mkPrimTyCon = do
           tcName <- coreToName tyConName tyConUnique qualifiedNameString tc
@@ -320,7 +358,7 @@ makeTyCon tc = tycon
 makeAlgTyConRhs :: AlgTyConRhs
                 -> C2C (Maybe C.AlgTyConRhs)
 makeAlgTyConRhs algTcRhs = case algTcRhs of
-  DataTyCon dcs _ _ -> Just <$> C.DataTyCon <$> mapM coreToDataCon dcs
+  DataTyCon {data_cons = dcs} -> Just <$> C.DataTyCon <$> mapM coreToDataCon dcs
   SumTyCon dcs _ -> Just <$> C.DataTyCon <$> mapM coreToDataCon dcs
 
 #if MIN_VERSION_ghc(8,10,0)
@@ -409,6 +447,8 @@ coreToTerm primMap unlocs = term
         go "Clash.Magic.noDeDup" args
           | [_aTy,f] <- args
           = C.Tick C.NoDeDup <$> term f
+        go "Clash.Magic.clashSimulation" _
+          = C.Data <$> coreToDataCon falseDataCon
         go "Clash.XException.xToErrorCtx" args
           -- xToErrorCtx :: forall a. String -> a -> a
           | [_ty, _msg, x] <- args
@@ -432,6 +472,15 @@ coreToTerm primMap unlocs = term
       Text.readMaybe (Text.unpack nm2)
 
     term' (Var x)                 = var x
+#if MIN_VERSION_ghc(8,8,0)
+    term' (Lit l@LitRubbish{}) = do
+      ty <- coreToType (literalType l)
+      return (C.Prim (C.PrimInfo (pack "_RUBBISH_")
+                                 ty
+                                 C.WorkNever
+                                 C.SingleResult
+                                 C.NoUnfolding))
+#endif
     term' (Lit l)                 = return $ C.Literal (coreToLiteral l)
     term' (App eFun (Type tyArg)) = C.TyApp <$> term eFun <*> coreToType tyArg
     term' (App eFun eArg)         = C.App   <$> term eFun <*> term eArg
@@ -510,7 +559,10 @@ coreToTerm primMap unlocs = term
           -> C.Cast <$> term e <*> coreToType ty1 <*> coreToType ty2
         _ -> term e
     term' (Tick (SourceNote rsp _) e) =
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+      C.Tick (C.SrcSpan (RealSrcSpan rsp GHC.Nothing)) <$>
+             addUsefull (RealSrcSpan rsp GHC.Nothing) (term e)
+#elif MIN_VERSION_ghc(9,0,0)
       C.Tick (C.SrcSpan (RealSrcSpan rsp Nothing)) <$>
              addUsefull (RealSrcSpan rsp Nothing) (term e)
 #else
@@ -554,15 +606,14 @@ coreToTerm primMap unlocs = term
                           CoreUnfolding {} -> do
                             sp <- view srcSpan
                             RWS.censor (const (SrcSpanRB sp)) (term (unfoldingTemplate unfolding))
-                          NoUnfolding -> error ("No unfolding for DC wrapper: " ++ showPpr unsafeGlobalDynFlags x)
-                          _ -> error ("Unexpected unfolding for DC wrapper: " ++ showPpr unsafeGlobalDynFlags x)
+                          NoUnfolding -> error ("No unfolding for DC wrapper: " ++ showPprUnsafe x)
+                          _ -> error ("Unexpected unfolding for DC wrapper: " ++ showPprUnsafe x)
               else C.Data <$> coreToDataCon dc
           Nothing -> case lookupPrim xNameS of
             Just (Just (Primitive f wi _))
               | Just n <- parseBundle "bundle" f        -> return (bundleUnbundleTerm (n+1) xType)
               | Just n <- parseBundle "unbundle" f      -> return (bundleUnbundleTerm (n+1) xType)
               | f == "Clash.Signal.Internal.mapSignal#" -> return (mapSignalTerm xType)
-              | f == "Clash.Signal.Internal.mapSignal#" -> return (mapSignalTerm xType)
               | f == "Clash.Signal.Internal.signal#"    -> return (signalTerm xType)
               | f == "Clash.Signal.Internal.appSignal#" -> return (appSignalTerm xType)
               | f == "Clash.Signal.Internal.traverse#"  -> return (traverseTerm xType)
@@ -609,9 +660,15 @@ coreToTerm primMap unlocs = term
               | otherwise
               -> C.Var <$> coreToId x
 
+#if MIN_VERSION_ghc(9,2,0)
+    alt _   (Alt DEFAULT      _  e) = (C.DefaultPat,) <$> term e
+    alt _   (Alt (LitAlt l)   _  e) = (C.LitPat (coreToLiteral l),) <$> term e
+    alt sp0 (Alt (DataAlt dc) xs e) = case span isTyVar xs of
+#else
     alt _   (DEFAULT   , _ , e) = (C.DefaultPat,) <$> term e
     alt _   (LitAlt l  , _ , e) = (C.LitPat (coreToLiteral l),) <$> term e
     alt sp0 (DataAlt dc, xs, e) = case span isTyVar xs of
+#endif
       (tyvs,tmvs) -> do
         (e',sp1) <- termSP sp0 e
         (,) <$> (C.DataPat <$> coreToDataCon dc
@@ -625,7 +682,7 @@ coreToTerm primMap unlocs = term
 #if MIN_VERSION_ghc(8,8,0)
       LitString  fs  -> C.StringLiteral (Char8.unpack fs)
       LitChar    c   -> C.CharLiteral c
-      LitRubbish     ->
+      LitRubbish{}   ->
         error $ "coreToTerm: Encountered LibRubbish. This is a bug in Clash. "
              ++ "Report on https://github.com/clash-lang/clash-compiler/issues."
 #else
@@ -637,12 +694,24 @@ coreToTerm primMap unlocs = term
 #else
       LitNumber lt i _ -> case lt of
 #endif
+#if MIN_VERSION_ghc(9,4,0)
+        LitNumBigNat  -> C.ByteArrayLiteral (ByteArray (integerToBigNatClamp# i))
+#else
         LitNumInteger -> C.IntegerLiteral i
         LitNumNatural -> C.NaturalLiteral i
+#endif
         LitNumInt     -> C.IntLiteral i
         LitNumInt64   -> C.IntLiteral i
         LitNumWord    -> C.WordLiteral i
         LitNumWord64  -> C.WordLiteral i
+#if MIN_VERSION_ghc(9,2,0)
+        LitNumInt8    -> C.Int8Literal i
+        LitNumInt16   -> C.Int16Literal i
+        LitNumInt32   -> C.Int32Literal i
+        LitNumWord8   -> C.Word8Literal i
+        LitNumWord16  -> C.Word16Literal i
+        LitNumWord32  -> C.Word32Literal i
+#endif
 #if MIN_VERSION_ghc(8,8,0)
       LitFloat r    -> C.FloatLiteral . floatToWord $ fromRational r
       LitDouble r   -> C.DoubleLiteral . doubleToWord $ fromRational r
@@ -734,7 +803,11 @@ hasPrimCo (AxiomRuleCo _ coers) = do
   tcs <- catMaybes <$> mapM hasPrimCo coers
   return (listToMaybe tcs)
 
+#if MIN_VERSION_ghc(9,6,0)
+hasPrimCo (SelCo _ co) = hasPrimCo co
+#else
 hasPrimCo (NthCo _ _ co)  = hasPrimCo co
+#endif
 hasPrimCo (LRCo _ co)   = hasPrimCo co
 hasPrimCo (InstCo co _) = hasPrimCo co
 hasPrimCo (SubCo co)    = hasPrimCo co
@@ -753,7 +826,9 @@ coreToDataCon dc = do
     mkDc dcTy repTys
   where
     mkDc dcTy repTys = do
-#if MIN_VERSION_ghc(8,10,0)
+#if MIN_VERSION_ghc(9,6,0)
+      let decLabel = decodeUtf8 . bytesFS . field_label . flLabel
+#elif MIN_VERSION_ghc(8,10,0)
       let decLabel = decodeUtf8 . bytesFS . flLabel
 #else
       let decLabel = decodeUtf8 . fastStringToByteString . flLabel
@@ -782,7 +857,7 @@ coreToDataCon dc = do
 
 hsImplBangToBool :: HsImplBang -> C.DcStrictness
 hsImplBangToBool HsLazy = C.Lazy
-hsImplBangToBool HsStrict = C.Strict
+hsImplBangToBool HsStrict{} = C.Strict
 hsImplBangToBool HsUnpack{} = C.Strict
 
 typeConstructorToString
@@ -791,9 +866,6 @@ typeConstructorToString
 typeConstructorToString constructor =
    Text.unpack . C.nameOcc <$> coreToName tyConName tyConUnique qualifiedNameString constructor
 
-_ATTR_NAME :: String
-_ATTR_NAME = "Clash.Annotations.SynthesisAttributes.Attr"
-
 -- | Flatten a list type structure to a list of types.
 listTypeToListOfTypes :: Type -> [Type]
 -- TyConApp ': [kind, head, tail]
@@ -813,118 +885,112 @@ boolTypeToBool (TyConApp constructor _args) = do
     _ -> error $ "Expected boolean constructor, got:" ++ constructorName
 boolTypeToBool s =
   error $ unwords [ "Could not unpack given type to bool:"
-                  , showPpr unsafeGlobalDynFlags s ]
+                  , showPprUnsafe s ]
 
 -- | Returns string of (LitTy (StrTyLit s)) construction.
 tyLitToString :: Type -> String
 tyLitToString (LitTy (StrTyLit s)) = unpackFS s
 tyLitToString s = error $ unwords [ "Could not unpack given type to string:"
-                                  , showPpr unsafeGlobalDynFlags s ]
+                                  , showPprUnsafe s ]
+
+-- | Returns string in Text form of (LitTy (StrTyLit s)) construction.
+tyLitToText :: Type -> Text
+tyLitToText = Text.pack . tyLitToString
 
 -- | Returns integer of (LitTy (NumTyLit n)) construction.
 tyLitToInteger :: Type -> Integer
 tyLitToInteger (LitTy (NumTyLit n)) = n
 tyLitToInteger s = error $ unwords [ "Could not unpack given type to integer:"
-                                   , showPpr unsafeGlobalDynFlags s ]
+                                   , showPprUnsafe s ]
 
 -- | Try to interpret a Type as an Attr
-coreToAttr
-  :: Type
-  -> C2C C.Attr'
-coreToAttr t@(TyConApp ty args) = do
-  let key   = args !! 0
-  let value = args !! 1
-  name' <- typeConstructorToString ty
+coreToAttr :: Type -> C2C (Attr Text)
+coreToAttr t0@(TyConApp ty args) = do
+  name <- typeConstructorToString ty
   envs <- view famInstEnvs
+  let
+    -- XXX: This relies on 'value' not being evaluated if the constructor
+    --      doesn't have a second field.
+    key = args !! 1
+    value = args !! 2
+#if MIN_VERSION_ghc(9,4,0)
+  let Reduction _ key1 = normaliseType envs Nominal key
+      Reduction _ value1 = normaliseType envs Nominal value
+#else
   let (_,key1) = normaliseType envs Nominal key
       (_,value1) = normaliseType envs Nominal value
-  case name' of
-    "Clash.Annotations.SynthesisAttributes.StringAttr" ->
-        return $ C.StringAttr' (tyLitToString key1) (tyLitToString value1)
-    "Clash.Annotations.SynthesisAttributes.IntegerAttr" ->
-        return $ C.IntegerAttr' (tyLitToString key1) (tyLitToInteger value1)
-    "Clash.Annotations.SynthesisAttributes.BoolAttr" -> do
-        bool <- boolTypeToBool value1
-        return $ C.BoolAttr' (tyLitToString key1) bool
-    "Clash.Annotations.SynthesisAttributes.Attr" ->
-        return $ C.Attr' (tyLitToString key1)
-    _ ->
-        case coreView t of
-          Just t' -> coreToAttr t'
-          Nothing -> error $ unwords [ "Expected StringAttr, IntegerAttr, BoolAttr or Attr"
-                       , "constructor, got:" ++ name' ]
+#endif
+  if
+    | name == show 'StringAttr ->
+      return $ StringAttr (tyLitToText key1) (tyLitToText value1)
+    | name == show 'IntegerAttr ->
+      return $ IntegerAttr (tyLitToText key1) (tyLitToInteger value1)
+    | name == show 'BoolAttr -> do
+      bool <- boolTypeToBool value1
+      return $ BoolAttr (tyLitToText key1) bool
+    | name == show 'Attr ->
+      return $ Attr (tyLitToText key1)
+    | otherwise ->
+      case coreView t0 of
+        Just t1 -> coreToAttr t1
+        Nothing -> error $ [__i|Expected constructor of Attr, got #{name}|]
+coreToAttr t0 =
+  case coreView t0 of
+    Just t1 -> coreToAttr t1
+    Nothing -> error $ [__i|Expected constructor of Attr, got #{showPprUnsafe t0}|]
 
-coreToAttr t =
-  error $ unwords [ "Expected type constructor (TyConApp), but got:"
-                  , showPpr unsafeGlobalDynFlags t ]
+coreToAttrs' :: [Type] -> C2C [Attr Text]
+coreToAttrs' [k, a, attrs] = do
+  -- We expect three type arguments:
+  --
+  --  k: either @Attr@ or @[Attr]@
+  --  a: type being annotated
+  --  attrs: attribute or list of attributes
+  --
+  attrs1 <- tryList
+  attrs2 <- tryAttr
+  case attrs1 <|> attrs2 of
+    Just theseAttrs -> do
+      subAttrs <- coreToAttrs a
+      pure (theseAttrs <> subAttrs)
+    Nothing ->
+      error [__i|
+        Expected either an attribute or a list of attributes, got:
 
-coreToAttrs'
-    :: [Type]
-    -> C2C [C.Attr']
-coreToAttrs' [annotationType, realType, attributes] = allAttrs
+          #{showPprUnsafe k}
+      |]
  where
-  allAttrs = (++) <$> attrs <*> subAttrs
 
-  subAttrs =
-    coreToAttrs realType
+  isListTy = fmap (== show ''[]) . typeConstructorToString
+  isAttrTy = fmap (== show ''Attr) . typeConstructorToString
 
-  attrs =
-    case annotationType of
-      TyConApp ty [TyConApp ty' _args'] -> do
-        name'  <- typeConstructorToString ty
-        name'' <- typeConstructorToString ty'
+  tryList = case k of
+    TyConApp ty0 [TyConApp ty1 _] -> do
+      ifM
+        (andM [isListTy ty0, isAttrTy ty1])
+        (Just <$> traverse coreToAttr (listTypeToListOfTypes attrs))
+        (pure Nothing)
+    _ -> pure Nothing
 
-        let result | name' == "GHC.Types.[]" && name'' == _ATTR_NAME =
-                      -- List of attributes
-                      traverse coreToAttr (listTypeToListOfTypes attributes)
-
-                   | name' == "GHC.Types.[]" =
-                      -- List, but unknown types
-                      error $ $(curLoc) ++ unwords [ "Annotate expects an"
-                                                   , "Attr or a list of"
-                                                   , "Attr's, but got a list"
-                                                   , "of:", name'']
-                   | otherwise =
-                        -- Some unknown nested type
-                      error $ $(curLoc) ++ unwords [ "Annotate expects an"
-                                                   , "Attr or a list of"
-                                                   , "Attr's, but got:"
-                                                   , name' ]
-
-        result
-
-      TyConApp ty _args -> do
-        name' <- typeConstructorToString ty
-        if name' == _ATTR_NAME
-          then
-            -- Single annotation
-            sequence [coreToAttr attributes]
-          else do
-            -- Annotation to something we don't recognize (not a list,
-            -- nor an Attr)
-            tystr <- typeConstructorToString ty
-            error $ unwords [ "Annotate expects an Attr or a list of Attr's,"
-                            , "but got:", tystr ]
-      _ ->
-        error $ $(curLoc) ++ unwords [ "Expected TyConApp, not:"
-                                     , showPpr unsafeGlobalDynFlags annotationType]
+  tryAttr = case k of
+    TyConApp ty _ -> do
+      ifM
+        (isAttrTy ty)
+        (Just <$> sequence [coreToAttr attrs])
+        (pure Nothing)
+    _ -> pure Nothing
 
 coreToAttrs' illegal =
-  error $ "Expected list with three items (as Annotate has three arguments), but got: "
-      ++ show (map (showPpr unsafeGlobalDynFlags) illegal)
+  error $ "Unexpected type args to Annotate: " ++ show (map (showPprUnsafe) illegal)
 
 -- | If this type has an annotate type synonym, return list of attributes.
-coreToAttrs
-  :: Type
-  -> C2C [C.Attr']
+coreToAttrs :: Type -> C2C [Attr Text]
 coreToAttrs (TyConApp tycon kindsOrTypes) = do
   name' <- typeConstructorToString tycon
 
-  if name' == "Clash.Annotations.SynthesisAttributes.Annotate"
-    then
-      coreToAttrs' kindsOrTypes
-    else
-      return []
+  if name' == show ''Annotate
+  then coreToAttrs' kindsOrTypes
+  else return []
 
 coreToAttrs _ =
     return []
@@ -958,15 +1024,23 @@ coreToType'
   -> C2C C.Type
 coreToType' (TyVarTy tv) = C.VarTy <$> coreToTyVar tv
 coreToType' (TyConApp tc args)
+#if MIN_VERSION_ghc(9,6,0)
+  | Just (FunTy _ _ ty1 ty2) <- tyConAppFunTy_maybe tc args = C.mkFunTy <$> coreToType ty1 <*> coreToType ty2
+#else
   | isFunTyCon tc = foldl C.AppTy (C.ConstTy C.Arrow) <$> mapM coreToType args
+#endif
   | otherwise     = case expandSynTyCon_maybe tc args of
+#if MIN_VERSION_ghc(9,6,0)
+                      ExpandsSyn substs synTy remArgs -> do
+#else
                       Just (substs,synTy,remArgs) -> do
+#endif
                         let substs' = mkTvSubstPrs substs
                             synTy'  = substTy substs' synTy
                         foldl C.AppTy <$> coreToType synTy' <*> mapM coreToType remArgs
                       _ -> do
                         tcName <- coreToName tyConName tyConUnique qualifiedNameString tc
-                        tyConMap %= (C.extendUniqMap tcName tc)
+                        tyConMap %= (C.insert tcName tc)
                         C.mkTyConApp <$> (pure tcName) <*> mapM coreToType args
 #if MIN_VERSION_ghc(8,8,0)
 coreToType' (ForAllTy (Bndr tv _) ty)   = C.ForAllTy <$> coreToTyVar tv <*> coreToType ty
@@ -985,13 +1059,17 @@ coreToType' (FunTy ty1 ty2)             = C.mkFunTy <$> coreToType ty1 <*> coreT
 #endif
 coreToType' (LitTy tyLit)    = return $ C.LitTy (coreToTyLit tyLit)
 coreToType' (AppTy ty1 ty2)  = C.AppTy <$> coreToType ty1 <*> coreToType' ty2
-coreToType' t@(CastTy _ _)   = error ("Cannot handle CastTy " ++ showPpr unsafeGlobalDynFlags t)
-coreToType' t@(CoercionTy _) = error ("Cannot handle CoercionTy " ++ showPpr unsafeGlobalDynFlags t)
+coreToType' (CastTy t (Refl{})) = coreToType' t
+coreToType' t@(CastTy _ _)   = error ("Cannot handle CastTy " ++ showPprUnsafe t)
+coreToType' t@(CoercionTy _) = error ("Cannot handle CoercionTy " ++ showPprUnsafe t)
 
 coreToTyLit :: TyLit
             -> C.LitTy
 coreToTyLit (NumTyLit i) = C.NumTy (fromInteger i)
 coreToTyLit (StrTyLit s) = C.SymTy (unpackFS s)
+#if MIN_VERSION_ghc(9,2,0)
+coreToTyLit (CharTyLit c) = C.CharTy c
+#endif
 
 coreToTyVar :: TyVar
             -> C2C C.TyVar
@@ -1116,21 +1194,22 @@ bundleUnbundleTerm nTyVarsExpected = go []
 -- @
 mapSignalTerm :: C.Type
               -> C.Term
-mapSignalTerm (C.ForAllTy aTV (C.ForAllTy bTV (C.ForAllTy clkTV funTy))) =
-    C.TyLam aTV (
-    C.TyLam bTV (
-    C.TyLam clkTV (
-    C.Lam   fId (
-    C.Lam   xId (
-    C.App (C.Var fId) (C.Var xId))))))
-  where
-    (C.FunTy _ funTy'') = C.tyView funTy
-    (C.FunTy aTy bTy)   = C.tyView funTy''
-    fName = C.mkUnsafeSystemName "f" 0
-    xName = C.mkUnsafeSystemName "x" 1
-    fTy   = C.mkFunTy aTy bTy
-    fId   = C.mkLocalId fTy fName
-    xId   = C.mkLocalId aTy xName
+mapSignalTerm (C.ForAllTy aTV (C.ForAllTy bTV (C.ForAllTy clkTV funTy)))
+  | (C.FunTy _ funTy'') <- C.tyView funTy
+  , (C.FunTy aTy bTy)   <- C.tyView funTy''
+  = let
+      fName = C.mkUnsafeSystemName "f" 0
+      xName = C.mkUnsafeSystemName "x" 1
+      fTy   = C.mkFunTy aTy bTy
+      fId   = C.mkLocalId fTy fName
+      xId   = C.mkLocalId aTy xName
+    in
+      C.TyLam aTV (
+      C.TyLam bTV (
+      C.TyLam clkTV (
+      C.Lam   fId (
+      C.Lam   xId (
+      C.App (C.Var fId) (C.Var xId))))))
 
 mapSignalTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1143,15 +1222,16 @@ mapSignalTerm ty = error $ $(curLoc) ++ show ty
 -- @/\(a:*)./\(dom:Domain).\(x:Signal dom a).x@
 signalTerm :: C.Type
            -> C.Term
-signalTerm (C.ForAllTy aTV (C.ForAllTy domTV funTy)) =
-    C.TyLam aTV (
-    C.TyLam domTV (
-    C.Lam   xId (
-    C.Var   xId)))
-  where
-    (C.FunTy _ saTy) = C.tyView funTy
-    xName = C.mkUnsafeSystemName "x" 0
-    xId   = C.mkLocalId saTy xName
+signalTerm (C.ForAllTy aTV (C.ForAllTy domTV funTy))
+  | (C.FunTy _ saTy) <- C.tyView funTy
+  = let
+      xName = C.mkUnsafeSystemName "x" 0
+      xId   = C.mkLocalId saTy xName
+    in
+      C.TyLam aTV (
+      C.TyLam domTV (
+      C.Lam   xId (
+      C.Var   xId)))
 
 signalTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1170,21 +1250,22 @@ signalTerm ty = error $ $(curLoc) ++ show ty
 -- @
 appSignalTerm :: C.Type
               -> C.Term
-appSignalTerm (C.ForAllTy domTV (C.ForAllTy aTV (C.ForAllTy bTV funTy))) =
-    C.TyLam domTV (
-    C.TyLam aTV (
-    C.TyLam bTV (
-    C.Lam   fId (
-    C.Lam   xId (
-    C.App (C.Var fId) (C.Var xId))))))
-  where
-    (C.FunTy _ funTy'') = C.tyView funTy
-    (C.FunTy saTy sbTy)   = C.tyView funTy''
-    fName = C.mkUnsafeSystemName "f" 0
-    xName = C.mkUnsafeSystemName "x" 1
-    fTy   = C.mkFunTy saTy sbTy
-    fId   = C.mkLocalId fTy fName
-    xId   = C.mkLocalId saTy xName
+appSignalTerm (C.ForAllTy domTV (C.ForAllTy aTV (C.ForAllTy bTV funTy)))
+  | (C.FunTy _ funTy'') <- C.tyView funTy
+  , (C.FunTy saTy sbTy) <- C.tyView funTy''
+  = let
+      fName = C.mkUnsafeSystemName "f" 0
+      xName = C.mkUnsafeSystemName "x" 1
+      fTy   = C.mkFunTy saTy sbTy
+      fId   = C.mkLocalId fTy fName
+      xId   = C.mkLocalId saTy xName
+    in
+      C.TyLam domTV (
+      C.TyLam aTV (
+      C.TyLam bTV (
+      C.Lam   fId (
+      C.Lam   xId (
+      C.App (C.Var fId) (C.Var xId))))))
 
 appSignalTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1202,16 +1283,17 @@ appSignalTerm ty = error $ $(curLoc) ++ show ty
 -- @
 vecUnwrapTerm :: C.Type
               -> C.Term
-vecUnwrapTerm (C.ForAllTy tTV (C.ForAllTy nTV (C.ForAllTy aTV funTy))) =
-    C.TyLam tTV (
-    C.TyLam nTV (
-    C.TyLam aTV (
-    C.Lam   vsId (
-    C.Var vsId))))
-  where
-    (C.FunTy _ vsTy) = C.tyView funTy
-    vsName           = C.mkUnsafeSystemName "vs" 0
-    vsId             = C.mkLocalId vsTy vsName
+vecUnwrapTerm (C.ForAllTy tTV (C.ForAllTy nTV (C.ForAllTy aTV funTy)))
+  | (C.FunTy _ vsTy) <- C.tyView funTy
+  = let
+        vsName = C.mkUnsafeSystemName "vs" 0
+        vsId   = C.mkLocalId vsTy vsName
+    in
+        C.TyLam tTV (
+        C.TyLam nTV (
+        C.TyLam aTV (
+        C.Lam   vsId (
+        C.Var vsId))))
 
 vecUnwrapTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1230,25 +1312,26 @@ vecUnwrapTerm ty = error $ $(curLoc) ++ show ty
 -- @
 traverseTerm :: C.Type
              -> C.Term
-traverseTerm (C.ForAllTy fTV (C.ForAllTy aTV (C.ForAllTy bTV (C.ForAllTy domTV funTy)))) =
-    C.TyLam fTV (
-    C.TyLam aTV (
-    C.TyLam bTV (
-    C.TyLam domTV (
-    C.Lam   dictId (
-    C.Lam   gId (
-    C.Lam   xId (
-    C.App (C.Var gId) (C.Var xId))))))))
-  where
-    (C.FunTy dictTy funTy1) = C.tyView funTy
-    (C.FunTy gTy    funTy2) = C.tyView funTy1
-    (C.FunTy xTy    _)      = C.tyView funTy2
-    dictName = C.mkUnsafeSystemName "dict" 0
-    gName    = C.mkUnsafeSystemName "g" 1
-    xName    = C.mkUnsafeSystemName "x" 2
-    dictId   = C.mkLocalId dictTy dictName
-    gId      = C.mkLocalId gTy gName
-    xId      = C.mkLocalId xTy xName
+traverseTerm (C.ForAllTy fTV (C.ForAllTy aTV (C.ForAllTy bTV (C.ForAllTy domTV funTy))))
+    | (C.FunTy dictTy funTy1) <- C.tyView funTy
+    , (C.FunTy gTy    funTy2) <- C.tyView funTy1
+    , (C.FunTy xTy    _)      <- C.tyView funTy2
+    = let
+        dictName = C.mkUnsafeSystemName "dict" 0
+        gName    = C.mkUnsafeSystemName "g" 1
+        xName    = C.mkUnsafeSystemName "x" 2
+        dictId   = C.mkLocalId dictTy dictName
+        gId      = C.mkLocalId gTy gName
+        xId      = C.mkLocalId xTy xName
+      in
+        C.TyLam fTV (
+        C.TyLam aTV (
+        C.TyLam bTV (
+        C.TyLam domTV (
+        C.Lam   dictId (
+        C.Lam   gId (
+        C.Lam   xId (
+        C.App (C.Var gId) (C.Var xId))))))))
 
 traverseTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1267,22 +1350,42 @@ traverseTerm ty = error $ $(curLoc) ++ show ty
 -- @/\(r:Rep)/\(a:TYPE Lifted)./\(b:TYPE r).\(f : (a -> b)).\(x : a).f x@
 dollarTerm :: C.Type
            -> C.Term
-dollarTerm (C.ForAllTy rTV (C.ForAllTy aTV (C.ForAllTy bTV funTy))) =
-    C.TyLam rTV (
-    C.TyLam aTV (
-    C.TyLam bTV (
-    C.Lam   fId (
-    C.Lam   xId (
-    C.App (C.Var fId) (C.Var xId))))))
-  where
-    (C.FunTy fTy funTy'') = C.tyView funTy
-    (C.FunTy aTy _)       = C.tyView funTy''
-    fName = C.mkUnsafeSystemName "f" 0
-    xName = C.mkUnsafeSystemName "x" 1
-    fId   = C.mkLocalId fTy fName
-    xId   = C.mkLocalId aTy xName
+#if MIN_VERSION_ghc(9,8,0)
+dollarTerm (C.ForAllTy raTV (C.ForAllTy rbTV (C.ForAllTy aTV (C.ForAllTy bTV funTy))))
+  | (C.FunTy fTy funTy'') <- C.tyView funTy
+  , (C.FunTy aTy _)       <- C.tyView funTy''
+  = let
+      fName = C.mkUnsafeSystemName "f" 0
+      xName = C.mkUnsafeSystemName "x" 1
+      fId   = C.mkLocalId fTy fName
+      xId   = C.mkLocalId aTy xName
+    in
+      C.TyLam raTV (
+      C.TyLam rbTV (
+      C.TyLam aTV (
+      C.TyLam bTV (
+      C.Lam   fId (
+      C.Lam   xId (
+      C.App (C.Var fId) (C.Var xId)))))))
+#else
+dollarTerm (C.ForAllTy rTV (C.ForAllTy aTV (C.ForAllTy bTV funTy)))
+  | (C.FunTy fTy funTy'') <- C.tyView funTy
+  , (C.FunTy aTy _)       <- C.tyView funTy''
+  = let
+      fName = C.mkUnsafeSystemName "f" 0
+      xName = C.mkUnsafeSystemName "x" 1
+      fId   = C.mkLocalId fTy fName
+      xId   = C.mkLocalId aTy xName
+    in
+      C.TyLam rTV (
+      C.TyLam aTV (
+      C.TyLam bTV (
+      C.Lam   fId (
+      C.Lam   xId (
+      C.App (C.Var fId) (C.Var xId))))))
+#endif
 
-dollarTerm ty = error $ $(curLoc) ++ show ty
+dollarTerm ty = error $ $(curLoc) ++ C.showPpr ty
 
 -- | Given the type:
 --
@@ -1306,17 +1409,18 @@ joinTerm ty = error $ $(curLoc) ++ show ty
 withFrozenCallStackTerm
   :: C.Type
   -> C.Term
-withFrozenCallStackTerm (C.ForAllTy aTV funTy) =
-  C.TyLam  aTV (
-  C.Lam    callStackId (
-  C.Lam    fId (
-  C.App (C.Var fId) (C.Var callStackId))))
-  where
-    (C.FunTy callStackTy fTy) = C.tyView funTy
-    callStackName = C.mkUnsafeSystemName "callStack" 0
-    fName         = C.mkUnsafeSystemName "f" 1
-    callStackId   = C.mkLocalId callStackTy callStackName
-    fId           = C.mkLocalId fTy fName
+withFrozenCallStackTerm (C.ForAllTy aTV funTy)
+  | (C.FunTy callStackTy fTy) <- C.tyView funTy
+  = let
+      callStackName = C.mkUnsafeSystemName "callStack" 0
+      fName         = C.mkUnsafeSystemName "f" 1
+      callStackId   = C.mkLocalId callStackTy callStackName
+      fId           = C.mkLocalId fTy fName
+    in
+      C.TyLam  aTV (
+      C.Lam    callStackId (
+      C.Lam    fId (
+      C.App (C.Var fId) (C.Var callStackId))))
 
 withFrozenCallStackTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1330,14 +1434,15 @@ withFrozenCallStackTerm ty = error $ $(curLoc) ++ show ty
 idTerm
   :: C.Type
   -> C.Term
-idTerm (C.ForAllTy aTV funTy) =
-  C.TyLam aTV (
-  C.Lam   xId (
-  C.Var xId))
-  where
-    (C.FunTy xTy _) = C.tyView funTy
-    xName           = C.mkUnsafeSystemName "x" 0
-    xId             = C.mkLocalId xTy xName
+idTerm (C.ForAllTy aTV funTy)
+  | (C.FunTy xTy _) <- C.tyView funTy
+  = let
+      xName           = C.mkUnsafeSystemName "x" 0
+      xId             = C.mkLocalId xTy xName
+    in
+      C.TyLam aTV (
+      C.Lam   xId (
+      C.Var xId))
 
 idTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1351,17 +1456,19 @@ idTerm ty = error $ $(curLoc) ++ show ty
 runRWTerm
   :: C.Type
   -> C.Term
-runRWTerm (C.ForAllTy rTV (C.ForAllTy oTV funTy)) =
-  C.TyLam rTV (
-  C.TyLam oTV (
-  C.Lam   fId (
-  (C.App (C.Var fId) (C.Prim (C.PrimInfo rwNm rwTy C.WorkNever C.SingleResult C.NoUnfolding))))))
- where
-    (C.FunTy fTy _)  = C.tyView funTy
-    (C.FunTy rwTy _) = C.tyView fTy
-    fName            = C.mkUnsafeSystemName "f" 0
-    fId              = C.mkLocalId fTy fName
-    rwNm             = pack "GHC.Prim.realWorld#"
+runRWTerm (C.ForAllTy rTV (C.ForAllTy oTV funTy))
+  | (C.FunTy fTy _)  <- C.tyView funTy
+  , (C.FunTy rwTy _) <- C.tyView fTy
+  = let
+      fName            = C.mkUnsafeSystemName "f" 0
+      fId              = C.mkLocalId fTy fName
+      rwNm             = pack "GHC.Prim.realWorld#"
+    in
+      C.TyLam rTV (
+      C.TyLam oTV (
+      C.Lam   fId (
+      (C.App (C.Var fId)
+             (C.Prim (C.PrimInfo rwNm rwTy C.WorkNever C.SingleResult C.NoUnfolding))))))
 
 runRWTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1375,23 +1482,24 @@ runRWTerm ty = error $ $(curLoc) ++ show ty
 checkUnpackUndefTerm
   :: C.Type
   -> C.Term
-checkUnpackUndefTerm (C.ForAllTy nTV (C.ForAllTy aTV funTy)) =
-  C.TyLam nTV (
-  C.TyLam aTV (
-  C.Lam knId (
-  C.Lam tpId (
-  C.Lam fId (
-  C.Var fId)))))
-  where
-    C.FunTy knTy r0Ty = C.tyView funTy
-    C.FunTy tpTy r1Ty = C.tyView r0Ty
-    C.FunTy fTy _     = C.tyView r1Ty
-    knName            = C.mkUnsafeSystemName "kn" 0
-    tpName            = C.mkUnsafeSystemName "tp" 1
-    fName             = C.mkUnsafeSystemName "f" 2
-    knId              = C.mkLocalId knTy knName
-    tpId              = C.mkLocalId tpTy tpName
-    fId               = C.mkLocalId fTy fName
+checkUnpackUndefTerm (C.ForAllTy nTV (C.ForAllTy aTV funTy))
+  | C.FunTy knTy r0Ty <- C.tyView funTy
+  , C.FunTy tpTy r1Ty <- C.tyView r0Ty
+  , C.FunTy fTy _     <- C.tyView r1Ty
+  = let
+      knName            = C.mkUnsafeSystemName "kn" 0
+      tpName            = C.mkUnsafeSystemName "tp" 1
+      fName             = C.mkUnsafeSystemName "f" 2
+      knId              = C.mkLocalId knTy knName
+      tpId              = C.mkLocalId tpTy tpName
+      fId               = C.mkLocalId fTy fName
+    in
+      C.TyLam nTV (
+      C.TyLam aTV (
+      C.Lam knId (
+      C.Lam tpId (
+      C.Lam fId (
+      C.Var fId)))))
 
 checkUnpackUndefTerm ty = error $ $(curLoc) ++ show ty
 
@@ -1406,17 +1514,18 @@ nameModTerm
   :: C.NameMod
   -> C.Type
   -> C.Term
-nameModTerm sa (C.ForAllTy nmTV (C.ForAllTy aTV funTy)) =
-  C.TyLam nmTV (
-  C.TyLam aTV (
-  C.Lam   xId (
-  (C.Tick (C.NameMod sa (C.VarTy nmTV)) (C.Var xId)))))
-  where
-    (C.FunTy xTy _)  = C.tyView funTy
-    -- Safe to use `mkUnsafeSystemName` here, because we're building the
-    -- identity \x.x, so any shadowing of 'x' would be the desired behavior.
-    xName            = C.mkUnsafeSystemName "x" 0
-    xId              = C.mkLocalId xTy xName
+nameModTerm sa (C.ForAllTy nmTV (C.ForAllTy aTV funTy))
+  | (C.FunTy xTy _) <- C.tyView funTy
+  = let
+      -- Safe to use `mkUnsafeSystemName` here, because we're building the
+      -- identity \x.x, so any shadowing of 'x' would be the desired behavior.
+      xName            = C.mkUnsafeSystemName "x" 0
+      xId              = C.mkLocalId xTy xName
+    in
+      C.TyLam nmTV (
+      C.TyLam aTV (
+      C.Lam   xId (
+      (C.Tick (C.NameMod sa (C.VarTy nmTV)) (C.Var xId)))))
 
 nameModTerm _ ty = error $ $(curLoc) ++ show ty
 
@@ -1431,20 +1540,21 @@ nameModTerm _ ty = error $ $(curLoc) ++ show ty
 xToErrorCtxTerm
   :: C.Type
   -> C.Term
-xToErrorCtxTerm (C.ForAllTy aTV funTy) =
-  C.TyLam aTV (
-  C.Lam ctxId (
-  C.Lam xId (
-  C.Var xId)))
-  where
-    (C.FunTy ctxTy rTy) = C.tyView funTy
-    (C.FunTy xTy _)     = C.tyView rTy
-    -- Safe to use `mkUnsafeSystemName` here, because we're building the
-    -- identity \_ x.x, so any shadowing of 'x' would be the desired behavior.
-    ctxName = C.mkUnsafeSystemName "ctx" 0
-    ctxId   = C.mkLocalId ctxTy ctxName
-    xName   = C.mkUnsafeSystemName "x" 1
-    xId     = C.mkLocalId xTy xName
+xToErrorCtxTerm (C.ForAllTy aTV funTy)
+  | (C.FunTy ctxTy rTy) <- C.tyView funTy
+  , (C.FunTy xTy _)     <- C.tyView rTy
+  = let
+      -- Safe to use `mkUnsafeSystemName` here, because we're building the
+      -- identity \_ x.x, so any shadowing of 'x' would be the desired behavior.
+      ctxName = C.mkUnsafeSystemName "ctx" 0
+      ctxId   = C.mkLocalId ctxTy ctxName
+      xName   = C.mkUnsafeSystemName "x" 1
+      xId     = C.mkLocalId xTy xName
+    in
+      C.TyLam aTV (
+      C.Lam ctxId (
+      C.Lam xId (
+      C.Var xId)))
 
 xToErrorCtxTerm ty = error $ $(curLoc) ++ show ty
 
