@@ -16,23 +16,30 @@ module Clash.GHC.GenerateBindings
   (generateBindings)
 where
 
-import           Control.Arrow           ((***), first)
+import           Control.Arrow           ((***))
 import           Control.DeepSeq         (deepseq)
 import           Control.Lens            ((%~),(&),(.~))
 import           Control.Monad           (unless)
 import qualified Control.Monad.State     as State
 import qualified Control.Monad.RWS.Strict as RWS
 import           Data.Coerce             (coerce)
-import           Data.Either             (partitionEithers, lefts, rights)
+import           Data.Either             (partitionEithers, lefts ,rights)
 import           Data.IntMap.Strict      (IntMap)
 import qualified Data.IntMap.Strict      as IMS
 import qualified Data.HashMap.Strict     as HashMap
 import           Data.List               (isPrefixOf)
+import           Data.Maybe              (listToMaybe)
 import qualified Data.Text               as Text
 import qualified Data.Time.Clock         as Clock
 
 import qualified GHC                     as GHC (Ghc)
 #if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+import qualified GHC.Types.SourceText    as GHC
+#endif
+#if MIN_VERSION_ghc(9,2,0)
+import qualified GHC.Utils.Panic         as GHC
+#endif
 import qualified GHC.Types.Basic         as GHC
 import qualified GHC.Core                as GHC
 import qualified GHC.Types.Demand        as GHC
@@ -63,6 +70,7 @@ import qualified TysWiredIn              as GHC
 import qualified Var                     as GHC
 import qualified SrcLoc                  as GHC
 #endif
+import           GHC.BasicTypes.Extra (isOpaque)
 
 import           Clash.Annotations.BitRepresentation.Internal (buildCustomReprs)
 import           Clash.Annotations.Primitive (HDL, extractPrim)
@@ -77,6 +85,7 @@ import           Clash.Core.Var          (Var (..), Id, IdScope (..), setIdScope
 import           Clash.Core.VarEnv
   (InScopeSet, VarEnv, emptyInScopeSet, extendInScopeSet, mkInScopeSet
   ,mkVarEnv, unionVarEnv, elemVarSet, mkVarSet)
+import qualified Clash.Data.UniqMap as UniqMap
 import           Clash.Debug             (traceIf)
 import           Clash.Driver            (compilePrimitive)
 import           Clash.Driver.Types      (BindingMap, Binding(..), IsPrim(..), ClashEnv(..), ClashDesign(..), ClashOpts(..))
@@ -89,8 +98,6 @@ import           Clash.Netlist.Types     (TopEntityT(..))
 import           Clash.Primitives.Types
   (Primitive (..), CompiledPrimMap)
 import           Clash.Primitives.Util   (generatePrimMap)
-import           Clash.Unique
-  (listToUniqMap, lookupUniqMap, mapUniqMap, unionUniqMap, uniqMapToUniqSet)
 import           Clash.Util              (reportTimeDiff)
 import qualified Clash.Util.Interpolate as I
 
@@ -140,13 +147,21 @@ generateBindings opts startAction primDirs importDirs dbs hdl modName dflagsM =
                    emptyGHC2CoreState
       (tcMap',tupTcCache)           = mkTupTyCons tcMap
       tcCache                       = makeAllTyCons tcMap' fiEnvs
-      allTcCache                    = tysPrimMap `unionUniqMap` tcCache
-      inScope0 = mkInScopeSet (uniqMapToUniqSet
-                      ((mapUniqMap (coerce . bindingId) bindingsMap) `unionUniqMap`
-                       (mapUniqMap (coerce . bindingId) clsMap)))
+      allTcCache                    = tysPrimMap <> tcCache
+      inScope0 = mkInScopeSet (
+                      (fmap (coerce . bindingId) bindingsMap) <>
+                      (fmap (coerce . bindingId) clsMap))
                                     -- Recursion info is always False for class
                                     -- selectors, no need to check free vars.
-      clsMap                        = mapUniqMap (\(v,i) -> (Binding v GHC.noSrcSpan GHC.Inline IsFun (mkClassSelector inScope0 allTcCache (varType v) i) False)) clsVMap
+      clsMap =
+        fmap (\(v,i) ->
+#if MIN_VERSION_ghc(9,4,0)
+               (Binding v GHC.noSrcSpan (GHC.Inline GHC.NoSourceText) IsFun
+#else
+               (Binding v GHC.noSrcSpan GHC.Inline IsFun
+#endif
+                  (mkClassSelector inScope0 allTcCache (varType v) i) False))
+             clsVMap
       allBindings                   = bindingsMap `unionVarEnv` clsMap
       topEntities'                  =
         (\m -> fst (RWS.evalRWS m (GHC2CoreEnv GHC.noSrcSpan fiEnvs) tcMap')) $
@@ -155,7 +170,7 @@ generateBindings opts startAction primDirs importDirs dbs hdl modName dflagsM =
             return (topEnt', annM, isTb)) topEntities
       topEntities'' =
         map (\(topEnt, annM, isTb) ->
-                case lookupUniqMap topEnt allBindings of
+                case UniqMap.lookup topEnt allBindings of
                   Just b -> TopEntityT (bindingId b) annM isTb
                   Nothing -> GHC.pgmError [I.i|
                     No top entity called '#{topEnt}' found. Make sure you are
@@ -176,10 +191,10 @@ generateBindings opts startAction primDirs importDirs dbs hdl modName dflagsM =
         , envTupleTyCons = tupTcCache
         , envPrimitives = primMapC
         , envCustomReprs = buildCustomReprs customBitRepresentations
+        , envDomains = domainConfs
         }
     , ClashDesign
         { designEntities = topEntities''
-        , designDomains = domainConfs
         , designBindings = allBindings'
         }
     )
@@ -194,7 +209,12 @@ setNoInlineTopEntities bm tes =
   ids = mkVarSet (fmap topId tes)
 
   go b@Binding{bindingId}
-    | bindingId `elemVarSet` ids = b { bindingSpec = GHC.NoInline }
+    | bindingId `elemVarSet` ids
+#if MIN_VERSION_ghc(9,4,0)
+    = b { bindingSpec = GHC.Opaque GHC.NoSourceText }
+#else
+    = b { bindingSpec = GHC.NoInline }
+#endif
     | otherwise = b
 
 -- TODO This function should be changed to provide the information that
@@ -296,10 +316,22 @@ checkPrimitive primMap v = do
       let
         info = GHC.idInfo v
         inline = GHC.inlinePragmaSpec $ GHC.inlinePragInfo info
+#if MIN_VERSION_ghc(9,4,0)
+        strictness = GHC.dmdSigInfo info
+#else
         strictness = GHC.strictnessInfo info
+#endif
         ty = GHC.varType v
+#if MIN_VERSION_ghc(9,2,0)
+        (argTys,_resTy) = GHC.splitFunTys (snd (GHC.splitForAllTyCoVars ty))
+#else
         (argTys,_resTy) = GHC.splitFunTys . snd . GHC.splitForAllTys $ ty
+#endif
+#if MIN_VERSION_ghc(9,4,0)
+        (dmdArgs,_dmdRes) = GHC.splitDmdSig strictness
+#else
         (dmdArgs,_dmdRes) = GHC.splitStrictSig strictness
+#endif
         nrOfArgs = length argTys
         loc = case GHC.getSrcLoc v of
                 GHC.UnhelpfulLoc _ -> ""
@@ -327,10 +359,16 @@ checkPrimitive primMap v = do
             warnArgs xs
 
       unless (qName == "Clash.XException.errorX" || "GHC." `isPrefixOf` qName) $ do
-        warnIf (inline /= GHC.NoInline)
+        warnIf (not (isOpaque inline))
+#if MIN_VERSION_ghc(9,4,0)
+          (primStr ++ "isn't marked OPAQUE."
+#else
           (primStr ++ "isn't marked NOINLINE."
+#endif
           ++ "\nThis might make Clash ignore this primitive.")
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,2,0)
+        warnIf (GHC.isDeadEndAppSig strictness nrOfArgs)
+#elif MIN_VERSION_ghc(9,0,0)
         warnIf (GHC.appIsDeadEnd strictness nrOfArgs)
 #else
         warnIf (GHC.appIsBottom strictness nrOfArgs)
@@ -353,25 +391,26 @@ mkClassSelector
   -> Term
 mkClassSelector inScope0 tcm ty sel = newExpr
   where
-    ((tvs,dictTy:_),_) = first (lefts *** rights)
-                       $ first (span (\l -> case l of Left _ -> True
-                                                      _      -> False))
-                       $ splitFunForallTy ty
-    newExpr = case tyView dictTy of
-      (TyConApp tcNm _)
-        | Just tc <- lookupUniqMap tcNm tcm
+    -- TODO: why can't we just use partitionEithers here?
+    (tvs,dicts) = (lefts *** rights)
+                . span (\l -> case l of {Left _ -> True; _ -> False})
+                $ fst (splitFunForallTy ty)
+    newExpr = case listToMaybe dicts of
+      Just dictTy@(tyView -> TyConApp tcNm _)
+        | Just tc <- UniqMap.lookup tcNm tcm
         , not (isNewTypeTc tc)
         -> flip State.evalState (0 :: Int) $ do
-                          dcId <- mkInternalVar inScope0 "dict" dictTy
-                          let inScope1 = extendInScopeSet inScope0 dcId
-                          selE <- mkSelectorCase "mkClassSelector" inScope1 tcm (Var dcId) 1 sel
-                          return (mkTyLams (mkLams selE [dcId]) tvs)
-      (FunTy arg res) -> flip State.evalState (0 :: Int) $ do
-                           dcId <- mkInternalVar inScope0 "dict" (mkFunTy arg res)
-                           return (mkTyLams (mkLams (Var dcId) [dcId]) tvs)
-      _ -> flip State.evalState (0 :: Int) $ do
-                           dcId <- mkInternalVar inScope0 "dict" dictTy
-                           return (mkTyLams (mkLams (Var dcId) [dcId]) tvs)
+              dcId <- mkInternalVar inScope0 "dict" dictTy
+              let inScope1 = extendInScopeSet inScope0 dcId
+              selE <- mkSelectorCase "mkClassSelector" inScope1 tcm (Var dcId) 1 sel
+              return (mkTyLams (mkLams selE [dcId]) tvs)
+      Just (tyView -> FunTy arg res) -> flip State.evalState (0 :: Int) $ do
+              dcId <- mkInternalVar inScope0 "dict" (mkFunTy arg res)
+              return (mkTyLams (mkLams (Var dcId) [dcId]) tvs)
+      Just dictTy -> flip State.evalState (0 :: Int) $ do
+              dcId <- mkInternalVar inScope0 "dict" dictTy
+              return (mkTyLams (mkLams (Var dcId) [dcId]) tvs)
+      Nothing -> error "mkClassSelector: expected at least one dictionary argument"
 
 mkTupTyCons :: GHC2CoreState -> (GHC2CoreState,IntMap TyConName)
 mkTupTyCons tcMap = (tcMap'',tupTcCache)
@@ -384,5 +423,5 @@ mkTupTyCons tcMap = (tcMap'',tupTcCache)
                  (GHC2CoreEnv GHC.noSrcSpan GHC.emptyFamInstEnvs)
                  tcMap
     tupTcCache       = IMS.fromList (zip [2..GHC.mAX_TUPLE_SIZE] (drop 3 tcNames))
-    tupHM            = listToUniqMap (zip tcNames tupTyCons)
-    tcMap''          = tcMap' & tyConMap %~ (`unionUniqMap` tupHM)
+    tupHM            = UniqMap.fromList (zip tcNames tupTyCons)
+    tcMap''          = tcMap' & tyConMap %~ (<> tupHM)
