@@ -44,7 +44,14 @@ import           Data.Set                    (Set)
 import qualified Data.Set                    as Set
 import           Data.Word                   (Word8)
 
--- GHC API
+-- GHC
+#if MIN_VERSION_ghc(9,8,0)
+import GHC.Types.Error (defaultOpts)
+import GHC.Iface.Errors.Ppr (missingInterfaceErrorDiagnostic)
+#endif
+#if MIN_VERSION_ghc(9,4,0)
+import           GHC.Driver.Env.KnotVars (emptyKnotVars)
+#endif
 #if MIN_VERSION_ghc(9,0,0)
 import           GHC.Types.Annotations (Annotation(..))
 import qualified GHC.Types.Annotations as Annotations
@@ -52,7 +59,6 @@ import qualified GHC.Core.Class as Class
 import qualified GHC.Core.FVs as CoreFVs
 import qualified GHC.Core as CoreSyn
 import qualified GHC.Types.Demand as Demand
-import           GHC.Driver.Session as DynFlags (unsafeGlobalDynFlags)
 import qualified GHC
 import qualified GHC.Types.Id as Id
 import qualified GHC.Types.Id.Info as IdInfo
@@ -65,7 +71,7 @@ import qualified GHC.Unit.Module.Env as ModuleEnv
 import qualified GHC.Utils.Monad as MonadUtils
 import qualified GHC.Types.Name as Name
 import qualified GHC.Types.Name.Env as NameEnv
-import           GHC.Utils.Outputable as Outputable (showPpr, showSDoc, text)
+import           GHC.Utils.Outputable as Outputable (text)
 import qualified GHC.Plugins as GhcPlugins (deserializeWithData, fromSerialized)
 import qualified GHC.IfaceToCore as TcIface
 import qualified GHC.Tc.Utils.Monad as TcRnMonad
@@ -80,7 +86,6 @@ import qualified Class
 import qualified CoreFVs
 import qualified CoreSyn
 import qualified Demand
-import           DynFlags                    (unsafeGlobalDynFlags)
 import qualified GHC
 import qualified Id
 import qualified IdInfo
@@ -91,7 +96,7 @@ import qualified MkCore
 import qualified Module
 import qualified MonadUtils
 import qualified Name
-import           Outputable                  (showPpr, showSDoc, text)
+import           Outputable                  (text)
 import qualified GhcPlugins                  (deserializeWithData, fromSerialized)
 import qualified TcIface
 import qualified TcRnMonad
@@ -111,6 +116,7 @@ import           Clash.Primitives.Util               (decodeOrErrJson, decodeOrE
 import           Clash.GHC.GHC2Core                  (qualifiedNameString')
 import           Clash.Util                          (curLoc)
 import qualified Clash.Util.Interpolate              as I
+import           Clash.GHC.Util
 
 -- | Data structure tracking loaded binders (and their related data)
 data LoadedBinders = LoadedBinders
@@ -154,11 +160,11 @@ bndrsInExpr e = partitionEithers (map go freeVars)
       Nothing -> error [I.i|
         Internal error: couldn't find class method
 
-          #{showPpr DynFlags.unsafeGlobalDynFlags v}
+          #{showPprUnsafe v}
 
         in class
 
-          #{showPpr DynFlags.unsafeGlobalDynFlags c}
+          #{showPprUnsafe c}
       |]
       Just n -> n
 
@@ -230,7 +236,11 @@ runIfl modName action = do
 
     globalEnv = TcRnTypes.IfGblEnv
       { TcRnTypes.if_doc = text "Clash.runIfl"
+#if MIN_VERSION_ghc(9,4,0)
+      , TcRnTypes.if_rec_types = emptyKnotVars
+#else
       , TcRnTypes.if_rec_types = Nothing
+#endif
       }
 
   hscEnv <- GHC.getSession
@@ -240,21 +250,34 @@ runIfl modName action = do
 loadDecl :: IfaceSyn.IfaceDecl -> TcRnTypes.IfL GHC.TyThing
 loadDecl = TcIface.tcIfaceDecl False
 
+#if MIN_VERSION_ghc(9,4,0)
+loadIface :: GHC.HscEnv -> GHC.Module -> IO (Maybe GHC.ModIface)
+loadIface env foundMod = do
+#else
 loadIface :: GHC.Module -> TcRnTypes.IfL (Maybe GHC.ModIface)
 loadIface foundMod = do
+#endif
+#if MIN_VERSION_ghc(9,4,0)
+  ifaceFailM <- LoadIface.findAndReadIface env (Outputable.text "loadIface")
+                  (fst (Module.getModuleInstantiation foundMod)) foundMod UnitTypes.NotBoot
+#elif MIN_VERSION_ghc(9,0,0)
   ifaceFailM <- LoadIface.findAndReadIface (Outputable.text "loadIface")
-#if MIN_VERSION_ghc(9,0,0)
                   (fst (Module.getModuleInstantiation foundMod)) foundMod UnitTypes.NotBoot
 #else
+  ifaceFailM <- LoadIface.findAndReadIface (Outputable.text "loadIface")
                   (fst (Module.splitModuleInsts foundMod)) foundMod False
 #endif
   case ifaceFailM of
     Maybes.Succeeded (modInfo,_) -> return (Just modInfo)
     Maybes.Failed msg -> let msg' = concat [ $(curLoc)
                                            , "Failed to load interface for module: "
-                                           , showPpr unsafeGlobalDynFlags foundMod
+                                           , showPprUnsafe foundMod
                                            , "\nReason: "
-                                           , showSDoc unsafeGlobalDynFlags msg
+#if MIN_VERSION_ghc(9,8,0)
+                                           , showSDocUnsafe (missingInterfaceErrorDiagnostic defaultOpts msg)
+#else
+                                           , showSDocUnsafe msg
+#endif
                                            ]
                          in traceIf True msg' (return Nothing)
 
@@ -300,7 +323,12 @@ getIfaceDeclM hdl bndr = do
     case Map.lookup nameMod lbCache of
       Nothing -> do
         -- Not loaded before
+#if MIN_VERSION_ghc(9,4,0)
+        env <- lift GHC.getSession
+        ifaceM <- lift (liftIO (loadIface env nameMod))
+#else
         ifaceM <- lift (runIfl nameMod (loadIface nameMod))
+#endif
         case ifaceM of
           Just iface -> do
             -- Add binder : decl map to cache
@@ -362,7 +390,10 @@ loadCustomReprAnnotations anns =
   env = Annotations.mkAnnEnv anns
   deserialize = GhcPlugins.deserializeWithData :: [Word8] -> DataReprAnn
 
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+  (mEnv, nEnv) = Annotations.deserializeAnns deserialize env
+  reprs = ModuleEnv.moduleEnvElts mEnv <> NameEnv.nonDetNameEnvElts nEnv
+#elif MIN_VERSION_ghc(9,0,0)
   (mEnv, nEnv) = Annotations.deserializeAnns deserialize env
   reprs = ModuleEnv.moduleEnvElts mEnv <> NameEnv.nameEnvElts nEnv
 #else
@@ -380,7 +411,7 @@ loadCustomReprAnnotations anns =
     error [I.i|
       Multiple DataReprAnn annotations for same type:
 
-        #{Outputable.showPpr DynFlags.unsafeGlobalDynFlags name}
+        #{showPprUnsafe name}
 
       Reprs:
 
@@ -464,20 +495,26 @@ loadExprFromTyThing :: CoreSyn.CoreBndr -> GHC.TyThing -> Maybe CoreSyn.CoreExpr
 loadExprFromTyThing bndr tyThing = case tyThing of
   GHC.AnId _id | Var.isId _id ->
     let _idInfo    = Var.idInfo _id
+#if MIN_VERSION_ghc(9,4,0)
+        unfolding  = IdInfo.realUnfoldingInfo _idInfo
+#else
         unfolding  = IdInfo.unfoldingInfo _idInfo
+#endif
     in case unfolding of
       CoreSyn.CoreUnfolding {} ->
         Just (CoreSyn.unfoldingTemplate unfolding)
       CoreSyn.DFunUnfolding dfbndrs dc es ->
         Just (MkCore.mkCoreLams dfbndrs (MkCore.mkCoreConApps dc es))
       CoreSyn.NoUnfolding
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+        | Demand.isDeadEndSig $ IdInfo.dmdSigInfo _idInfo
+#elif MIN_VERSION_ghc(9,0,0)
         | Demand.isDeadEndSig $ IdInfo.strictnessInfo _idInfo
 #else
         | Demand.isBottomingSig $ IdInfo.strictnessInfo _idInfo
 #endif
         -> do
-          let noUnfoldingErr = "no_unfolding " ++ showPpr unsafeGlobalDynFlags bndr
+          let noUnfoldingErr = "no_unfolding " ++ showPprUnsafe bndr
           Just (MkCore.mkAbsentErrorApp (Var.varType _id) noUnfoldingErr)
       _ -> Nothing
   _ -> Nothing
