@@ -2,7 +2,7 @@
   Copyright   :  (C) 2013-2016, University of Twente,
                      2016-2017, Myrtle Software Ltd,
                      2017     , Google Inc.
-                     2021     , QBayLogic B.V.
+                     2021-2023, QBayLogic B.V.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 -}
@@ -29,16 +29,17 @@ where
 #endif
 
 -- External Modules
-import           Clash.Annotations.Primitive     (HDL, PrimitiveGuard)
+import           Clash.Annotations.Primitive     (HDL, PrimitiveGuard(..))
 import           Clash.Annotations.TopEntity     (TopEntity (..))
 import           Clash.Primitives.Types          (UnresolvedPrimitive)
 import           Clash.Util                      (ClashException(..), pkgIdFromTypeable)
 import qualified Clash.Util.Interpolate          as I
 import           Control.Arrow                   (first)
 import           Control.Exception               (SomeException, throw)
-import           Control.Monad                   (forM, when)
+import           Control.Monad                   (forM, join, when)
 import           Data.List.Extra                 (nubSort)
 import           Control.Exception               (throwIO)
+import           Control.Monad                   (foldM)
 #if MIN_VERSION_ghc(9,0,0)
 import           Control.Monad.Catch             as MC (try)
 #endif
@@ -53,7 +54,8 @@ import qualified Data.HashMap.Strict             as HashMap
 import           Data.Typeable                   (Typeable)
 import           Data.List                       (foldl', nub, find)
 import qualified Data.Map                        as Map
-import           Data.Maybe                      (catMaybes, fromMaybe, mapMaybe)
+import           Data.Maybe
+  (catMaybes, fromMaybe, listToMaybe, mapMaybe)
 import qualified Data.Text                       as Text
 import qualified Data.Text.Encoding              as Text
 import qualified Data.Time.Clock                 as Clock
@@ -63,6 +65,9 @@ import           Debug.Trace
 import           Language.Haskell.TH.Syntax      (lift)
 import           GHC.Natural                     (naturalFromInteger)
 import           GHC.Stack                       (HasCallStack)
+#if MIN_VERSION_ghc(9,4,0)
+import           System.FilePath.Posix           (dropExtension, takeDirectory)
+#endif
 import           Text.Read                       (readMaybe)
 
 #ifdef USE_GHC_PATHS
@@ -77,6 +82,43 @@ import           System.Process                  (runInteractiveCommand,
 
 -- GHC API
 #if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+import           GHC.Driver.Phases (StopPhase(NoStop))
+import           GHC.Driver.Pipeline (mkPipeEnv, runPipeline, hscBackendPipeline)
+#if MIN_VERSION_ghc(9,6,0)
+import           GHC.SysTools.Cpp (offsetIncludePaths)
+import           GHC.Unit.Home.ModInfo (homeMod_bytecode)
+#else
+import           GHC.Driver.Pipeline.Execute (offsetIncludePaths)
+import           GHC.Driver.Pipeline.Monad (PipelineOutput(NoOutputFile, Persistent))
+#endif
+import           GHC.Driver.Pipeline.Monad ( MonadUse(use) )
+import           GHC.Driver.Pipeline.Phases (TPhase(T_HscPostTc))
+import           GHC.Data.Bool (OverridingBool)
+import           GHC.Driver.Config.Tidy (initTidyOpts)
+import           GHC.Driver.Errors.Types (GhcMessage(GhcTcRnMessage))
+import           GHC.Driver.Monad (modifySession)
+import           GHC.Unit.Env (addHomeModInfoToHug)
+import           GHC.Unit.Home.ModInfo (HomeModInfo(HomeModInfo))
+import           GHC.Unit.Module.ModSummary (findTarget)
+#else
+import           GHC.Utils.Misc (OverridingBool)
+#endif
+#if MIN_VERSION_ghc(9,2,0)
+import qualified GHC.Driver.Env as HscTypes
+import qualified GHC.Unit.Module.ModGuts as HscTypes
+import qualified GHC.Types.SourceError as HscTypes
+import qualified GHC.Unit.Module.Deps as HscTypes
+import qualified GHC.Driver.Backend as Backend
+import qualified GHC.Unit.Module.Graph as Graph
+import qualified GHC.Platform.Ways as Ways
+#if !MIN_VERSION_ghc(9,4,0)
+import qualified GHC.Types.Error as Error
+#endif
+#else
+import qualified GHC.Driver.Types as HscTypes
+import qualified GHC.Driver.Ways as Ways
+#endif
 import qualified GHC.Types.Annotations as Annotations
 import qualified GHC.Core.FVs as CoreFVs
 import qualified GHC.Core as CoreSyn
@@ -88,7 +130,6 @@ import qualified GHC.Driver.Session as DynFlags
 import qualified GHC.Data.FastString as FastString
 import qualified GHC
 import qualified GHC.Driver.Main as HscMain
-import qualified GHC.Driver.Types as HscTypes
 import qualified GHC.Utils.Monad as MonadUtils
 import qualified GHC.Utils.Panic as Panic
 import qualified GHC.Serialized as Serialized (deserializeWithData)
@@ -107,9 +148,7 @@ import qualified GHC.Types.Name.Occurrence as OccName
 import           GHC.Utils.Outputable (ppr)
 import qualified GHC.Utils.Outputable as Outputable
 import qualified GHC.Types.Unique.Set as UniqSet
-import           GHC.Utils.Misc (OverridingBool)
 import qualified GHC.Types.Var as Var
-import qualified GHC.Driver.Ways as Ways
 import qualified GHC.Unit.Module.Env as ModuleEnv
 import qualified GHC.Types.Name.Env as NameEnv
 #else
@@ -235,7 +274,12 @@ setupGhc useColor dflagsM idirs = do
         df <- do
           df <- GHC.getSessionDynFlags
 #if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,2,0)
+          logger <- GHC.getLogger
+          df1 <- liftIO (GHC.interpretPackageEnv logger df)
+#else
           df1 <- liftIO (GHC.interpretPackageEnv df)
+#endif
           _ <- GHC.setSessionDynFlags df1
 
 #else
@@ -256,9 +300,21 @@ setupGhc useColor dflagsM idirs = do
         return dfPlug
 
   let dflags1 = dflags
-                  { DynFlags.optLevel = 2
-                  , DynFlags.ghcMode  = GHC.CompManager
+                  { DynFlags.ghcMode  = GHC.CompManager
                   , DynFlags.ghcLink  = GHC.LinkInMemory
+#if !MIN_VERSION_ghc(9,4,0)
+                  , DynFlags.optLevel = 2
+#endif
+#if MIN_VERSION_ghc(9,2,0)
+                  , DynFlags.backend  =
+                      if Ways.hostIsProfiled
+#if MIN_VERSION_ghc(9,6,0)
+                         then Backend.noBackend
+#else
+                         then Backend.NoBackend
+#endif
+                         else Backend.platformDefaultBackend (DynFlags.targetPlatform dflags)
+#else
                   , DynFlags.hscTarget
 #if MIN_VERSION_ghc(9,0,0)
                       = if Ways.hostIsProfiled
@@ -271,6 +327,7 @@ setupGhc useColor dflagsM idirs = do
                                   DynFlags.targetPlatform
 #endif
                                     dflags
+#endif
                   , DynFlags.reductionDepth = 1000
                   }
   let dflags2 = unwantedOptimizationFlags dflags1
@@ -290,7 +347,12 @@ setupGhc useColor dflagsM idirs = do
                ])
       (return ())
 
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,2,0)
+  _ <- GHC.setSessionDynFlags dflags3
+  hscenv <- GHC.getSession
+  hscenv1 <- MonadUtils.liftIO (DynamicLoading.initializePlugins hscenv)
+  GHC.setSession hscenv1
+#elif MIN_VERSION_ghc(9,0,0)
   _ <- GHC.setSessionDynFlags dflags3
   hscenv <- GHC.getSession
   dflags4 <- MonadUtils.liftIO (DynamicLoading.initializePlugins hscenv dflags3)
@@ -318,13 +380,22 @@ loadLocalModule
        , [CoreSyn.CoreBind]                     -- All bindings
        )
 loadLocalModule hdl modName = do
+#if MIN_VERSION_ghc(9,4,0)
+  target <- GHC.guessTarget modName Nothing Nothing
+#else
   target <- GHC.guessTarget modName Nothing
+#endif
   GHC.setTargets [target]
   modGraph <- GHC.depanal [] False
   let modGraph' = GHC.mapMG disableOptimizationsFlags modGraph
       -- 'topSortModuleGraph' ensures that modGraph2, and hence tidiedMods
       -- are in topological order, i.e. the root module is last.
-      modGraph2 = Digraph.flattenSCCs (GHC.topSortModuleGraph True modGraph' Nothing)
+      modGraph2 = Digraph.flattenSCCs $
+#if MIN_VERSION_ghc(9,2,0)
+                  -- TODO: this might break backpack
+                  Graph.filterToposortToModules $
+#endif
+                  GHC.topSortModuleGraph True modGraph' Nothing
 
   liftIO $ mapM_ checkMonoLocalBindsMod modGraph2
 
@@ -350,12 +421,83 @@ loadLocalModule hdl modName = do
     --
     -- Given that TH splices can do non-trivial computation and I/O,
     -- running TH twice must be avoid.
+#if MIN_VERSION_ghc(9,4,0)
+    let (tc_result,_) = GHC.tm_internals_ tcMod
+    let tcMod' = tcMod
+#else
     tcMod' <- GHC.loadModule tcMod
+#endif
     dsMod <- fmap GHC.coreModule $ GHC.desugarModule tcMod'
     hsc_env <- GHC.getSession
     simpl_guts <- MonadUtils.liftIO $ HscMain.hscSimplify hsc_env [] dsMod
     checkForInvalidPrelude simpl_guts
+#if MIN_VERSION_ghc(9,4,0)
+    opts <- liftIO (initTidyOpts hsc_env)
+    (tidy_guts,_) <- MonadUtils.liftIO $ TidyPgm.tidyProgram opts simpl_guts
+#else
     (tidy_guts,_) <- MonadUtils.liftIO $ TidyPgm.tidyProgram hsc_env simpl_guts
+#endif
+#if MIN_VERSION_ghc(9,4,0)
+    let
+      loadAsByteCode
+        | Just GHC.Target { targetAllowObjCode = obj }
+            <- findTarget m (HscTypes.hsc_targets hsc_env)
+        , not obj
+        = True
+        | otherwise = False
+      lcl_dflags = GHC.ms_hspp_opts m
+      old_paths  = GHC.includePaths lcl_dflags
+      location = GHC.ms_location m
+      input_fn = fromMaybe (error "loadLocalModule") (GHC.ml_hs_file location)
+      basename = dropExtension input_fn
+      current_dir = takeDirectory basename
+#if MIN_VERSION_ghc(9,6,0)
+      interpreterBackend = Backend.interpreterBackend
+#else
+      interpreterBackend = Backend.Interpreter
+#endif
+      (bcknd, dflags3)
+        | loadAsByteCode
+        = ( interpreterBackend
+          , DynFlags.gopt_set
+              (lcl_dflags { GHC.backend = interpreterBackend })
+              Opt_ForceRecomp
+          )
+        | otherwise
+        = (GHC.backend dflags, lcl_dflags)
+      dflags = dflags3
+              { GHC.includePaths = offsetIncludePaths dflags3 $
+                                    DynFlags.addImplicitQuoteInclude
+                                      old_paths
+                                      [current_dir] }
+#if MIN_VERSION_ghc(9,6,0)
+      pipelineOutput = Backend.backendPipelineOutput bcknd
+#else
+      pipelineOutput = case bcknd of
+        GHC.Interpreter -> NoOutputFile
+        GHC.NoBackend -> NoOutputFile
+        _ -> Persistent
+#endif
+      upd_summary = m { GHC.ms_hspp_opts = dflags }
+      hsc_env1 = HscTypes.hscSetFlags dflags hsc_env
+      pipe_env = mkPipeEnv NoStop input_fn Nothing pipelineOutput
+      pipeline = do
+        ac <- use (T_HscPostTc hsc_env1 upd_summary
+                      (TcRnTypes.FrontendTypecheck tc_result) mempty Nothing )
+        hscBackendPipeline pipe_env hsc_env1 upd_summary ac
+    (iface, linkable) <- liftIO (runPipeline (HscTypes.hsc_hooks hsc_env1) pipeline)
+#if MIN_VERSION_ghc(9,6,0)
+    details <- liftIO (HscMain.initModDetails hsc_env1 iface)
+    linkable1 <- liftIO (traverse (HscMain.initWholeCoreBindings hsc_env1 iface details)
+                                  (homeMod_bytecode linkable))
+    let linkable2 = linkable {homeMod_bytecode = linkable1}
+#else
+    details <- liftIO (HscMain.initModDetails hsc_env1 upd_summary iface)
+    let linkable2 = linkable
+#endif
+    let mod_info = HomeModInfo iface details linkable2
+    modifySession $ HscTypes.hscUpdateHUG (addHomeModInfoToHug mod_info)
+#endif
     let pgm        = HscTypes.cg_binds tidy_guts
     let modFamInstEnv = TcRnTypes.tcg_fam_inst_env $ fst $ GHC.tm_internals_ tcMod
     _ <- GHC.setSessionDynFlags oldDFlags
@@ -376,7 +518,7 @@ loadLocalModule hdl modName = do
   localPrims <- findPrimitiveAnnotations hdl binderIds
   let loaded1 = loaded0{lbPrims=lbPrims loaded0 <> Seq.fromList localPrims}
 
-  let allBinders = concat binders ++ makeRecursiveGroups (Map.assocs (lbBinders loaded0))
+  let allBinders = makeRecursiveGroups (Map.assocs (lbBinders loaded0))
   pure (rootIds, modFamInstEnvs', rootModule, loaded1, allBinders)
 
 nameString :: Name.Name -> String
@@ -446,7 +588,15 @@ loadModules startAction useColor hdl modName dflagsM idirs = do
     famInstEnvs <- do
       (msgs, m) <- TcRnMonad.liftIO $ TcRnMonad.initTcInteractive hscEnv FamInst.tcGetFamInstEnvs
       case m of
-        Nothing -> TcRnMonad.liftIO $ throwIO (HscTypes.mkSrcErr (snd msgs))
+        Nothing -> TcRnMonad.liftIO $ throwIO
+                                    $ HscTypes.mkSrcErr
+#if MIN_VERSION_ghc(9,4,0)
+                                    $ fmap GhcTcRnMessage msgs
+#elif MIN_VERSION_ghc(9,2,0)
+                                    $ Error.getErrorMessages msgs
+#else
+                                    $ snd msgs
+#endif
         Just x  -> return x
 
     allSyn     <- Map.fromList <$> findSynthesizeAnnotations allBinderIds
@@ -530,7 +680,8 @@ loadModules startAction useColor hdl modName dflagsM idirs = do
 #endif
 
         famToDomain  = fromMaybe (error "KnownConf: Expected Symbol at LHS of type family")
-                         . fmap fsToText . Type.isStrLitTy . head . FamInstEnv.fi_tys
+                         . join . fmap (fmap fsToText) . fmap Type.isStrLitTy
+                         . listToMaybe . FamInstEnv.fi_tys
         famToConf    = unpackKnownConf . FamInstEnv.fi_rhs
 
         knownConfNms = fmap famToDomain knownConfs
@@ -636,25 +787,42 @@ errOnDuplicateAnnotations
   -> [[a]]
   -- ^ Parsed annotations
   -> [(CoreSyn.CoreBndr, a)]
-errOnDuplicateAnnotations nm bndrs anns =
+errOnDuplicateAnnotations nm =
+  combineAnnotationsWith err nm
+ where
+  err _ _ = Left $ "A binder can't have more than one '" ++ nm ++ "' annotation."
+
+combineAnnotationsWith
+  :: forall a. (a -> a -> Either String a)
+  -- ^ function to (attempts to) combine different annotations
+  -> String
+  -- ^ Name of annotation
+  -> [CoreSyn.CoreBndr]
+  -- ^ Binders searched for
+  -> [[a]]
+  -- ^ Parsed annotations
+  -> [(CoreSyn.CoreBndr, a)]
+combineAnnotationsWith f nm bndrs anns =
   go (zip bndrs anns)
  where
-  go
-    :: [(CoreSyn.CoreBndr, [a])]
-    -> [(CoreSyn.CoreBndr, a)]
+  go :: [(CoreSyn.CoreBndr, [a])] -> [(CoreSyn.CoreBndr, a)]
   go []             = []
   go ((_, []):ps)   = go ps
-  go ((b, [p]):ps)  = (b, p) : (go ps)
-  go ((b, _):_)  =
-    Panic.pgmError $ "The following value has multiple "
-                  ++ "'" ++ nm ++ "' annotations: "
-                  ++ Outputable.showSDocUnsafe (ppr b)
+  go ((b, (a:as)):ps) = case foldM f a as of
+    Left err ->
+      Panic.pgmError $ "Error processing '" ++ nm ++ "' annotations on "
+                       ++ Outputable.showSDocUnsafe (pprQualified $ Var.varName b)
+                       ++ ":\n" ++ err
+    Right x -> (b, x) : go ps
+  pprQualified :: Name.Name -> Outputable.SDoc
+  pprQualified x = case Name.nameModule_maybe x of
+    Just m  -> Outputable.hcat [ppr m, Outputable.dot, ppr x]
+    Nothing -> ppr x
+
 
 -- | Find annotations by given targets
 findAnnotationsByTargets
-  :: GHC.GhcMonad m
-  => Typeable a
-  => Data a
+  :: (GHC.GhcMonad m, Data a, Typeable a)
   => [Annotations.AnnTarget Name.Name]
   -> m [[a]]
 findAnnotationsByTargets targets =
@@ -666,15 +834,15 @@ findAnnotationsByTargets targets =
 
 -- | Find all annotations of a certain type in all modules seen so far.
 findAllModuleAnnotations
-  :: GHC.GhcMonad m
-  => Data a
-  => Typeable a
+  :: (GHC.GhcMonad m, Data a, Typeable a)
   => m [a]
 findAllModuleAnnotations = do
   hsc_env <- GHC.getSession
   ann_env <- liftIO $ HscTypes.prepareAnnotations hsc_env Nothing
   return $ concat
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+        $ (\(mEnv,nEnv) -> ModuleEnv.moduleEnvElts mEnv <> NameEnv.nonDetNameEnvElts nEnv)
+#elif MIN_VERSION_ghc(9,0,0)
          $ (\(mEnv,nEnv) -> ModuleEnv.moduleEnvElts mEnv <> NameEnv.nameEnvElts nEnv)
 #else
          $ UniqFM.nonDetEltsUFM
@@ -689,9 +857,7 @@ findAllModuleAnnotations = do
 
 -- | Find all annotations belonging to all binders seen so far.
 findNamedAnnotations
-  :: GHC.GhcMonad m
-  => Data a
-  => Typeable a
+  :: (GHC.GhcMonad m, Data a, Typeable a)
   => [CoreSyn.CoreBndr]
   -> m [[a]]
 findNamedAnnotations bndrs =
@@ -703,8 +869,14 @@ findPrimitiveGuardAnnotations
   -> m [(Text.Text, (PrimitiveGuard ()))]
 findPrimitiveGuardAnnotations bndrs = do
   anns0 <- findNamedAnnotations bndrs
-  let anns1 = errOnDuplicateAnnotations "PrimitiveGuard" bndrs anns0
+  let anns1 = combineAnnotationsWith combinePrimGuards "PrimitiveGuard" bndrs anns0
   pure (map (first (qualifiedNameString' . Var.varName)) anns1)
+ where
+  combinePrimGuards a b = case (a,b) of
+    (HasBlackBox x _, HasBlackBox y _) -> Right (HasBlackBox (x++y) ())
+    (DontTranslate  , DontTranslate)   -> Right DontTranslate
+    (_,_) -> Left "One binder can't have both HasBlackBox and DontTranslate annotations."
+
 
 -- | Find annotations of type @DataReprAnn@ and convert them to @DataRepr'@
 findCustomReprAnnotations
@@ -811,19 +983,29 @@ findPrimitiveAnnotations hdl bndrs = do
 
 parseModule :: GHC.GhcMonad m => GHC.ModSummary -> m GHC.ParsedModule
 parseModule modSum = do
+#if MIN_VERSION_ghc(9,2,0)
+  (GHC.ParsedModule pmModSum pmParsedSource extraSrc) <-
+    GHC.parseModule modSum
+  return (GHC.ParsedModule
+            (disableOptimizationsFlags pmModSum)
+            pmParsedSource extraSrc)
+#else
   (GHC.ParsedModule pmModSum pmParsedSource extraSrc anns) <-
     GHC.parseModule modSum
   return (GHC.ParsedModule
             (disableOptimizationsFlags pmModSum)
             pmParsedSource extraSrc anns)
+#endif
 
 disableOptimizationsFlags :: GHC.ModSummary -> GHC.ModSummary
 disableOptimizationsFlags ms@(GHC.ModSummary {..})
   = ms {GHC.ms_hspp_opts = dflags}
   where
     dflags = unwantedOptimizationFlags (ms_hspp_opts
-              { DynFlags.optLevel = 2
-              , DynFlags.reductionDepth = 1000
+              { DynFlags.reductionDepth = 1000
+#if !MIN_VERSION_ghc(9,4,0)
+              , DynFlags.optLevel = 2
+#endif
               })
 
 unwantedOptimizationFlags :: GHC.DynFlags -> GHC.DynFlags
@@ -944,20 +1126,40 @@ removeStrictnessAnnotations pm =
     rmTyClD tyClD = tyClD
 
     -- rmDataDefn :: GHC.DataId name => GHC.HsDataDefn name -> GHC.HsDataDefn name
+#if MIN_VERSION_ghc(9,2,0)
+    rmDataDefn :: GHC.HsDataDefn GHC.GhcPs -> GHC.HsDataDefn GHC.GhcPs
+#endif
     rmDataDefn hdf = hdf {GHC.dd_cons = (fmap . fmap) rmCD (GHC.dd_cons hdf)}
 
     -- rmCD :: GHC.DataId name => GHC.ConDecl name -> GHC.ConDecl name
     rmCD gadt@(GHC.ConDeclGADT {}) = gadt {GHC.con_res_ty = rmHsType (GHC.con_res_ty gadt)
+#if MIN_VERSION_ghc(9,2,0)
+                                          ,GHC.con_g_args = rmGConDetails (GHC.con_g_args gadt)
+#else
                                           ,GHC.con_args   = rmConDetails (GHC.con_args gadt)
+#endif
                                           }
     rmCD h98@(GHC.ConDeclH98 {})   = h98  {GHC.con_args = rmConDetails (GHC.con_args h98)}
 #if !MIN_VERSION_ghc(9,0,0)
     rmCD xcon                      = xcon
 #endif
 
+#if MIN_VERSION_ghc(9,4,0)
+    rmGConDetails :: GHC.HsConDeclGADTDetails GHC.GhcPs -> GHC.HsConDeclGADTDetails GHC.GhcPs
+    rmGConDetails (GHC.PrefixConGADT args) = GHC.PrefixConGADT (fmap rmHsScaledType args)
+    rmGConDetails (GHC.RecConGADT rec tkn) = GHC.RecConGADT ((fmap . fmap . fmap) rmConDeclF rec) tkn
+#elif MIN_VERSION_ghc(9,2,0)
+    rmGConDetails :: GHC.HsConDeclGADTDetails GHC.GhcPs -> GHC.HsConDeclGADTDetails GHC.GhcPs
+    rmGConDetails (GHC.PrefixConGADT args) = GHC.PrefixConGADT (fmap rmHsScaledType args)
+    rmGConDetails (GHC.RecConGADT rec)     = GHC.RecConGADT ((fmap . fmap . fmap) rmConDeclF rec)
+#endif
+
     -- type HsConDeclDetails name = HsConDetails (LBangType name) (Located [LConDeclField name])
     -- rmConDetails :: _ => GHC.HsConDeclDetails name -> GHC.HsConDeclDetails name
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,2,0)
+    rmConDetails (GHC.PrefixCon tys args) = GHC.PrefixCon tys (fmap rmHsScaledType args)
+    rmConDetails (GHC.InfixCon l r)   = GHC.InfixCon (rmHsScaledType l) (rmHsScaledType r)
+#elif MIN_VERSION_ghc(9,0,0)
     rmConDetails (GHC.PrefixCon args) = GHC.PrefixCon (fmap rmHsScaledType args)
     rmConDetails (GHC.InfixCon l r)   = GHC.InfixCon (rmHsScaledType l) (rmHsScaledType r)
 #else
@@ -970,12 +1172,22 @@ removeStrictnessAnnotations pm =
     -- rmHsType :: GHC.DataId name => GHC.Located (GHC.HsType name) -> GHC.Located (GHC.HsType name)
     rmHsType = transform go
       where
+#if MIN_VERSION_ghc(9,2,0)
+        go ::
+          GHC.LBangType GHC.GhcPs ->
+          GHC.LBangType GHC.GhcPs
+#endif
         go (GHC.unLoc -> GHC.HsBangTy _ _ ty) = ty
         go ty                               = ty
 
 #if MIN_VERSION_ghc(9,0,0)
     rmHsScaledType = transform go
       where
+#if MIN_VERSION_ghc(9,2,0)
+        go ::
+          GHC.HsScaled GHC.GhcPs (GHC.LBangType GHC.GhcPs) ->
+          GHC.HsScaled GHC.GhcPs (GHC.LBangType GHC.GhcPs)
+#endif
         go (GHC.HsScaled m (GHC.unLoc -> GHC.HsBangTy _ _ ty)) = GHC.HsScaled m ty
         go ty = ty
 #endif
@@ -996,8 +1208,14 @@ checkForInvalidPrelude guts =
     []    -> return ()
     (x:_) -> throw (ClashException noSrcSpan (msgWrongPrelude x) Nothing)
   where
+#if MIN_VERSION_ghc(9,4,0)
+    pkgs = HscTypes.dep_direct_pkgs . HscTypes.mg_deps $ guts
+#else
     pkgs = HscTypes.dep_pkgs . HscTypes.mg_deps $ guts
-#if MIN_VERSION_ghc(9,0,0)
+#endif
+#if MIN_VERSION_ghc(9,4,0)
+    pkgIds = map (UnitTypes.unitIdString) (toList pkgs)
+#elif MIN_VERSION_ghc(9,0,0)
     pkgIds = map (UnitTypes.unitIdString . fst) pkgs
 #else
     pkgIds = map (GhcPlugins.installedUnitIdString . fst) pkgs
